#
# $Copyright: Copyright (c) 2011 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $RCSfile: OraDGBrokerAgent.pm,v $ $Revision: 1.3.2.3 $
#

package OraDGBrokerAgent;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use ag_i18n_inc;

my $SU = "/bin/su";
my $PASSWDFILE = "/etc/passwd";

@EXPORT = qw(
);

$VERSION = '1.00';

require 5.000;

%default = {
	resname => '',

	VADCtlMode => '',

	# Elements to store the resource atributes
	# ========================================

	# Oracle instance ID used as OARCLE_SID when calling the dgmgrl
	Sid => '',

	# Oracle user ID used to SU for dgmgrl
	Owner => '',

	# Oracle home directory used for ORACLE_HOME
	Home => '',

	# Force failover even if remote cluster is still active but
	# remote RAC instance (service group) is offline
	ForceDB => '',

	# AgentDebug will enable exhaustive messages from every 
	# data guard broker CLI call
	AgentDebug => '',

	# Elements holding static information
	# ==============================================

	# The hostname as retrieved from the OS and used by Oracle
	hostname => '',

	# The nodename as used by VCS/VAD
	nodename => '',

	# The service group name for this resource
	grpname => '',

	# The name of the remote cluster
	remcls => '',

	# The Oracle command to call the data guard broker CLI
	dgmgrl => '',

	# Timeout / alarm value for the dgmgrl command default 20 sec
	# This does not apply to the failover / switchover calls!
	dgalarm => '',

	# Retry count for the dgmgrl show commands
	dgsretry => '',

	# Elements holding dynamic data
	#==============================

	# The state of the remote cluster
	remstate => '',

	# stdout lines from external commands called using backticks
	cmdout => '',

	# The unique name of the database as returned by dgmgrl
	# for the instance
	dbname => '',

	# The current intended role of the instance as reported by dgmgrl
	dbrole => '',

	# The errorr message from the last dgmgrl command
	orclerrmsg => ''
};

#
# Fill the global hash with all resource relevant information including
# values retrieved from the attribute list
#
# Return values:	pointer to global hash = success
# 			undef = something went wrong an the calling
# 			EP has to exit / return state unknown
sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;

	my $ret;
	my $self = { %default };

	$self->{'resname'} = shift;
	my @Args = @_;

	# Check for VAD controlmode return value if > 0
	my $clus_prod = $ENV{"CLUSTER_PRODUCT"};
	$clus_prod = "VCS" if (! defined($clus_prod));
	$self->{'VADCtlMode'} = 0;
	if ($clus_prod eq "VAD") {
		($ret, $self->{'VADCtlMode'}) = VCSAG_GET_ATTR_VALUE ("ControlMode", -1, 1, @ARGV);
		return undef if ($ret != $VCSAG_SUCCESS);
	}

	# Get Arguments

	($ret, $self->{'Sid'}) = VCSAG_GET_ATTR_VALUE ("Sid", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'Owner'}) = VCSAG_GET_ATTR_VALUE ("Owner", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	my $fnd = 0;
	open(PWF, "<", $PASSWDFILE);
	while (<PWF>) {
		next unless /^$self->{'Owner'}:/;
		$fnd = 1;
		last;
	}
	close PWF;
	unless ($fnd) {
		VCSAG_LOG_MSG("E", "User defined by attribute Owner does not exist in local password file", 501);
		return undef;
	}

	($ret, $self->{'Home'}) = VCSAG_GET_ATTR_VALUE ("Home", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'ForceDB'}) = VCSAG_GET_ATTR_VALUE ("ForceDB", -1, 1, @Args);
	if ($ret != $VCSAG_SUCCESS) {
		# ForceDB was added for Q4/2009 it may not be in the argument list
		$self->{'ForceDB'} = '';
	}

	($ret, $self->{'AgentDebug'}) = VCSAG_GET_ATTR_VALUE ("AgentDebug", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'Encoding'}) = VCSAG_GET_ATTR_VALUE ("Encoding", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	chomp($self->{'hostname'} = `uname -n`);

	$self->{'dgmgrl'} = "(ORACLE_SID=$self->{'Sid'}\;";
	$self->{'dgmgrl'} .= "ORACLE_HOME=$self->{'Home'}\;";
	$self->{'dgmgrl'} .= "export ORACLE_SID ORACLE_HOME\;";
	$self->{'dgmgrl'} .= "$SU $self->{'Owner'} -c \"";
	$self->{'dgmgrl'} .= "$self->{'Home'}/bin/dgmgrl / \\\"<DGMCMD>\\\"\")";

	if ($self->{'ForceDB'}) {
		$self->{'srvctl'} = "(ORACLE_HOME=$self->{'Home'}\;";
		$self->{'srvctl'} .= "export ORACLE_HOME\;";
		$self->{'srvctl'} .= "$self->{'Home'}/bin/srvctl status";
		$self->{'srvctl'} .= " database -d $self->{'ForceDB'})";
	}

	$self->{'dgalarm'} = 20;
	$self->{'dgsretry'} = 5;

	bless $self, $class;

	$self->get_local_nodename;

	return $self;
}

sub destroy {}

# Flush all child groups on the local node to reset faults of
# Oracle resources caused by Broker interaction
sub cg_flush {
	my $self = shift;

	my $cgrp = '';

	open(GDEP, "$HAGRP -dep $self->{'grpname'} 2>&1 |") or
		return 1;
	while (<GDEP>) {
		next if /^#/;
		(undef, $cgrp, undef) = split;
		system("$HAGRP -flush $cgrp -sys $self->{'nodename'} -localclus 2>&1 1>/dev/null");
	}
	close GDEP;
	return 0;
}

# Freeze / unfreeze all child groups to supress cluster interaction
# which contain a resource of type Oracle for the same Sid
# while the broker restarts instances during a switch / failover
#
# Return value 	0 - success for all groups
#		1 - error on open $HAGRP -dep
# 		>0 the accumalated return values of $HAGRP -[un]freeze
sub cg_freeze {
	my $self = shift;
	my $froper = shift;
	if ($froper) {
		$froper = "-freeze";
	}
	else {
		$froper = "-unfreeze";
	}
	my $cgrp = '';
	my $ret = 0;

	open(GDEP, "$HAGRP -dep $self->{'grpname'} 2>&1 |") or
		return 1;
	while (<GDEP>) {
		next if /^#/;
		(undef, $cgrp, undef) = split;
		my $gres = `$HAGRP -resources $cgrp 2>&1`;
		chomp $gres;
		foreach my $indres (split(/\n/, $gres)) {
			next unless (`$HARES -value $indres Type 2>&1` =~ /Oracle/);
			next unless (`$HARES -value $indres Sid 2>&1` =~ /$self->{'Sid'}/);
			$ret += system("$HAGRP $froper $cgrp -localclus 2>&1 1>/dev/null");
			last;
		}
	}
	close GDEP;
	return $ret;
}

# Execute a single data guard CLI command given as the second value in the
# call list
# 
# Return values:	the return value of the dgmgrl command plus
# 			stdout / stderr in $self->{'cmdout'}
sub exec_dgmgrl {
	my $self = shift;
	my $dgbcmd = shift;

	# Reset the cdmout variable since it is extended
	$self->{'cmdout'} = '';

	my $cmd = $self->{'dgmgrl'};
	$cmd =~ s/<DGMCMD>/$dgbcmd/;
 
	VCSAG_LOGDBG_MSG(1, "exec_dgmgrl running dgmgrl command :: $dgbcmd ");
	my $pid = open(DGCM, "$cmd 2>&1 |") or return 2;

	eval {
		local $SIG{ALRM} = sub { die "dgalarm timer exhausted" };
		alarm $self->{'dgalarm'};
		while (<DGCM>) {
			$self->{'cmdout'} .= $_;
		}
		close DGCM;
		alarm 0;
	};

	if ($@ and $@ =~ /dgalarm timer exhausted/) {
		kill 9, $pid;
		close DGCM;
		VCSAG_LOG_MSG("W", "dgalarm timer exhausted when calling $cmd", 523, $cmd) if $self->{'AgentDebug'};
		$self->{'cmdout'} .= "VCS $self->{'resname'} dgalarm timer = $self->{'dgalarm'} sec exhausted while executing dgmgrl command";
		return 1;
	}
	return 0;
}

# Execute a Cluster or Data Guard Broker command at a cluster site
# using an action EP
#
# Return values:	0 on sucessful execution on at least one node
# 				$self->{'cmdout'} will contain the action or
# 				command output
# 			1 - if execution does not complete on any remote node
sub exec_action {
	my $self = shift;
	my $action_args = shift;
	my ($clnode, $nname);

	# Call the ActRemote action entry point with the rest of
	# action_args as a string against each remote node
	open(RNOD, "$HARES -state $self->{'resname'} 2>&1 |") or
		return 1;
	while (<RNOD>) {
		next if /^#/;
		next if /\s+localclus\:/;
		(undef, undef, $clnode, undef) = split;
		(undef, $nname) = split(/:/, $clnode);
		next unless (defined $nname);
		$self->{'cmdout'} = `$HARES -action $self->{'resname'} ActRemote -actionargs $action_args -sys $nname -clus $self->{'remcls'} 2>&1`;
		return 0 unless $?;
	}
}

# Function to extend the timout value of an entry point
# useful for action and info with short timeout values
sub extend_ep_timeout {
	my $self = shift;
	my $tmo_val = shift;
	return 1 unless (defined $tmo_val);

	my $tmo_ext_file = "/var/VRTSvcs/log/tmp/" . "$self->{'resname'}" . ".tmo";
	open(TEF, "> $tmo_ext_file") or return 1;
	print TEF "$tmo_val";
	close TEF;
	return 0;
}

# Get group and remote cluster name for the resource
#
# Return values:	0 - success, $self->{'grpname'} & $self->{'remcls'}
#			1 - fault in retrieving information
sub get_clus_info {
	my $self = shift;

	$self->{'grpname'} = `$HARES -value $self->{'resname'} Group 2>&1`;
	return 1 if ($?);

	chomp ($self->{'grpname'});
	open(RCLUS, "$HAGRP -display $self->{'grpname'} -attribute Cluster 2>&1 |");
	while (<RCLUS>) {
		next if /^#/;
		chomp;
		(undef, undef, undef, $self->{'remcls'}) = split;
		last;
	}
	close RCLUS;
	return 1 if ($self->{'remcls'} eq '');

	return 0;
}

# Retrieve the name of the database associated by Oracle with the instance
# defined by the SID attribute
#
# Return values:	0 - success, $self->{'dbname'} is updated
# 			1 - database name can not be retrieved
sub get_db_name {
	my $self = shift;
	my $dgbcmd = "show instance '".$self->{'Sid'}."'";

	$self->exec_dgmgrl($dgbcmd);
	$self->{'dbname'} = '';

	foreach (split(/\n/, $self->{'cmdout'})) {
		chomp;
		if (s/^instance.*$self->{'Sid'}.*database\s+\'//i) {
			s/\'.*$//;
			$self->{'dbname'} = $_;
			next;
		}
		if (/ORA-\d+/) {
			$self->{'orclerrmsg'} = $_;
			VCSAG_LOG_MSG("W", "The show instance call returned Oracle message $self->{'orclerrmsg'}", 520, $self->{'orclerrmsg'}) if $self->{'AgentDebug'};
			last;
		}
		last if (/SUCCESS/i);
	}
	if ($self->{'dbname'} ne '') {
		return 0;
	}
	else {
		VCSAG_LOG_MSG("E", "Can not determine database name for instance $self->{'SID'}", 521, $self->{'SID'}) if $self->{'AgentDebug'};
		return 1;
	}
}

# Get the value of the "Role" field from the dgmgrl "show database" command
#
# Return values:	0 - success, $self->{'dbrole'} is updated
#			1 - the current role could not be retrieved
#			    for example Oracle might not be running
#
#			extensive message if AgentDebug is enabled
sub get_db_role {
	my $self = shift;
	my $instmatch = 0;

	my $dgbcmd = "show database '".$self->{'dbname'}."'";
	$self->exec_dgmgrl($dgbcmd);

	$self->{'dbrole'} = '';

	foreach (split(/\n/, $self->{'cmdout'})) {
		chomp;
		if (s/^\s+Role:\s+//i) {
			if (/PRIMARY/i) {
				$self->{'dbrole'} = $_;
			}
			else {
				$self->{'dbrole'} = $_;
			}
		}
		if (s/\s+$self->{'Sid'}//i) {
			$instmatch = 1;
			if (/apply instance/) {
				$self->{'dbrole'} .= " apply instance";
			}
		}
		if (/ORA-\d+/) {
			$self->{'orclerrmsg'} = $_;
			VCSAG_LOG_MSG("W", "The show instance call returned Oracle message $self->{'orclerrmsg'}", 520, $self->{'orclerrmsg'}) if $self->{'AgentDebug'};
			last;
		}
		last if (/SUCCESS/i);
	}
	if ($self->{'dbname'} ne '' && $instmatch == 1) {
		return 0;
	}
	else {
		VCSAG_LOG_MSG("E", "Can not determine role for database $self->{'dbname'}, $dgbcmd returned $self->{'cmdout'}", 522, $self->{'dbname'}, $dgbcmd,  $self->{'cmdout'}) if $self->{'AgentDebug'};
		return 1;
	}
}

# Function : get_local_nodename
# Desc     : Retrieve the name of the local node
# Returns  : 0 success - $self->{'nodename'} is updated

sub get_local_nodename {
	my $self = shift;
	my $SysName = "${VCSCONF}/conf/sysname";
	my $GlobZoneName = "$VCSCONF/.vcshost";
	my $LocSys;

	if ( -r "$SysName" ) {
		open(SNAME, "<$SysName");
	}
	elsif ( -r "$GlobZoneName" ) {
		open(SNAME, "<$GlobZoneName");
	}
	else {
		open(SNAME, "uname -n |");
	}
	chomp($self->{'nodename'} = <SNAME>);
	close SNAME;
	return 0;
}

# Get the state of the remote cluster for the resource.
# Only a two cluster = single standby configuration is supported!
#
# Return values:	0 - success, $self->{'remstate'} is updated
# 			1 - remote state can not be retrieved

sub get_remote_state {
	my $self = shift;

	$self->{'remstate'} = `$HACLUS -state $self->{'remcls'} 2>&1`;
	chomp($self->{'remstate'});

	return 0;
}

# Get the state of the remote RAC instances
#
# Return values:	0 - success, $self->{'cmdout'} is updated
# 			1 - remote state can not be retrieved

sub rac_state {
	my $self = shift;

	$self->{'cmdout'} = `$self->{'srvctl'} 2>&1`;
	print $self->{'cmdout'};

	return 0;
}

1;
